import unittest
import warnings
from pprint import pprint

from ds_foundation.handlers.abstract_handlers import ConnectorContract, HandlerFactory
from ds_foundation.managers.data_properties import DataPropertyManager


class S3HandlerTest(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_runs(self):
        """Basic smoke test"""
        pass

    def test_source_modified_and_list(self):
        connector_contract = ConnectorContract(resource='data/synthetic/synthetic_customer.csv', connector_type='dsv',
                                               location='discovery-connector-scratchpad',
                                               module_name='ds_connectors.handlers.aws_s3_handlers',
                                               handler='S3SourceHandler')
        handler = HandlerFactory.instantiate(connector_contract)
        result = handler.list_bucket(connector_contract.location, starts_with='data/synthetic/sy')
        control = ['data/synthetic/synthetic_agent.csv', 'data/synthetic/synthetic_customer.csv']
        self.assertEqual(control, result)
        result = handler.list_bucket(connector_contract.location, ends_with='er.csv')
        control = ['data/synthetic/synthetic_customer.csv']
        self.assertEqual(control, result)
        result = handler.get_modified()
        control = '2019-08-16 12:18:24+00:00'
        self.assertEqual(control, str(result))

    def test_load_dsv(self):
        connector_contract = ConnectorContract(resource='data/synthetic/synthetic_customer.csv', connector_type='dsv',
                                               location='ds-discovery',
                                               module_name='ds_connectors.handlers.aws_s3_handlers',
                                               handler='S3SourceHandler',
                                               kwargs={'delimiter': ',', 'encoding': 'latin1'})
        handler = HandlerFactory.instantiate(connector_contract)
        result = handler.load_canonical()
        control = ['surname', 'forename', 'gender', 'id', 'balance', 'age', 'start', 'profession', 'online', 'single num', 'weight_num', 'null', 'single cat', 'weight_cat', 'last_login', 'status']
        self.assertEqual(control, list(result))

    def test_load_pickle(self):
        def test_load_dsv(self):
            connector_contract = ConnectorContract(resource='data/synthetic/synthetic_customer.csv',
                                                   connector_type='dsv',
                                                   location='discovery-connector-scratchpad',
                                                   module_name='ds_connectors.handlers.aws_s3_handlers',
                                                   handler='S3SourceHandler',
                                                   kwargs={'delimiter': ',', 'encoding': 'latin1'})
            handler = HandlerFactory.instantiate(connector_contract)
            result = handler.load_canonical()
            control = ['surname', 'forename', 'gender', 'id', 'balance', 'age', 'start', 'profession', 'online',
                       'single num', 'weight_num', 'null', 'single cat', 'weight_cat', 'last_login', 'status']
            self.assertEqual(control, list(result))


if __name__ == '__main__':
    unittest.main()



from io import StringIO
from copy import deepcopy
from botocore.exceptions import ClientError
import boto3
try:
    import cPickel as pickle
except ImportError:
    import pickle

from ds_foundation.handlers.abstract_handlers import AbstractSourceHandler, ConnectorContract, AbstractPersistHandler
from ds_connectors.parsers.dsv import DelimitedParser

__author__ = 'Darryl Oatridge, Neil Pasricha'


class AwsS3SourceHandler(AbstractSourceHandler):
    """ An AWS S3 source handler

        Delimitor Separated Value (dsv):
            encoding: the source encoding format. defualt utf-8
            file_stream: a String IO file stream to parse
            fieldnames: a set of field names for the dictionary if no header
            restkey: If row has more fields than fieldnames, remaining data is put in a list under restkey
            restval: If non-blank row has fewer fields than fieldnames, values are filled-in with restval
            dialect: Dialect class grouping formatting parameters
            delimiter: A one-character string used to separate fields. It defaults to ','.
            quotechar: A one-character string used to quote fields containing special characters, such as the
                        delimiter or quotechar, or which contain new-line characters. It defaults to '"'.
            escapechar: A one-character string used by the writer to escape the delimiter if quoting is set to
                        QUOTE_NONE and the quotechar if doublequote is False.
            doublequote: Controls how instances of quotechar appearing inside a field should themselves be quoted.
                        When True, the character is doubled. Default
                        When False, the escapechar is used as a prefix to the quotechar
            skipinitialspace: When True, whitespace immediately following the delimiter is ignored. Default False
            lineterminator: The string used to terminate lines produced by the writer.
            quoting: Controls when quotes should be generated by the writer and recognised by the reader
    """

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Hander passing the connector_contract dictionary """
        super().__init__(connector_contract)
        self._modified = 0

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['csv', 'dsv', 'pickle']

    def load_canonical(self) -> dict:
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The S3 Source Connector Contract has not been set correctly")
        resource = self.connector_contract.resource
        connector_type = self.connector_contract.connector_type
        bucket = self.connector_contract.location
        _kwargs = self.connector_contract.kwargs
        if connector_type.lower() not in self.supported_types():
            raise ValueError(f"The source type '{connector_type}' is not supported. see supported_types()")
        s3_client = boto3.client('s3')
        try:
            s3_object = s3_client.get_object(Bucket=bucket, Key=resource)
        except ClientError as e:
            code = e.response["Error"]["Code"]
            raise ConnectionError(f"Failed to retrieve the object from S3 client with error code '{code}")
        _kwargs = deepcopy(self.connector_contract.kwargs)
        if connector_type.lower() in ['dsv', 'csv']:
            encoding = 'utf-8' if _kwargs.get('encoding') is None else _kwargs.pop('encoding')
            resource_body = s3_object['Body'].read().decode(encoding)
            return DelimitedParser.read_dsv(StringIO(resource_body), **_kwargs)
        if connector_type.lower() in ['pickle']:
            return pickle.load(s3_object['Body'].read())
        raise LookupError(f'The connector type {connector_type} is not currently supported')

    def get_modified(self) -> [int, float, str]:
        """ returns if the file has been modified"""
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The S3 Source Connector Contract has not been set correctly")
        resource = self.connector_contract.resource
        bucket = self.connector_contract.location
        s3_client = boto3.client('s3')
        try:
            s3_object = s3_client.get_object(Bucket=bucket, Key=resource)
        except ClientError as e:
            code = e.response["Error"]["Code"]
            raise ConnectionError(f"Failed to retrieve the object from S3 client with error code '{code}")
        return s3_object.get('LastModified')

    @staticmethod
    def list_bucket(bucket: str, starts_with: str=None, ends_with: str=None) -> list:
        """ Returns True is the file exists """
        s3_client = boto3.client('s3')
        response = s3_client.list_objects_v2(Bucket=bucket)
        rtn_list = []
        for obj in response.get('Contents', []):
            key = str(obj.get('Key'))
            if isinstance(starts_with, str) and not key.startswith(starts_with):
                continue
            if isinstance(ends_with, str) and not key.endswith(ends_with):
                continue
            rtn_list.append(key)
        return rtn_list


class AwsS3PersistHandler(AbstractPersistHandler):

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Hander passing the connector_contract dictionary """
        super().__init__(connector_contract)
        self._modified = 0

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['pickle']

    def get_modified(self) -> [int, float, str]:
        """ returns if the file has been modified"""
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The S3 Source Connector Contract has not been set correctly")
        resource = self.connector_contract.resource
        bucket = self.connector_contract.location
        s3_client = boto3.client('s3')
        try:
            s3_object = s3_client.get_object(Bucket=bucket, Key=resource)
        except ClientError as e:
            code = e.response["Error"]["Code"]
            raise ConnectionError(f"Failed to retrieve the object from S3 client with error code '{code}")
        return s3_object.get('LastModified')

    def exists(self) -> bool:
        """ Returns True is the file exists """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The S3 Source Connector Contract has not been set correctly")
        resource = self.connector_contract.resource
        bucket = self.connector_contract.location
        s3_client = boto3.client('s3')
        response = s3_client.list_objects_v2(Bucket=bucket)
        for obj in response.get('Contents', []):
            if obj['Key'] == resource:
                return True
        return False

    def load_canonical(self) -> dict:
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The S3 Source Connector Contract has not been set correctly")
        resource = self.connector_contract.resource
        connector_type = self.connector_contract.connector_type
        bucket = self.connector_contract.location
        _kwargs = self.connector_contract.kwargs
        if connector_type.lower() not in self.supported_types():
            raise ValueError(f"The source type '{connector_type}' is not supported. see supported_types()")
        s3_client = boto3.client('s3')
        try:
            s3_object = s3_client.get_object(Bucket=bucket, Key=resource)
        except ClientError as e:
            code = e.response["Error"]["Code"]
            raise ConnectionError(f"Failed to retrieve the object from S3 client with error code '{code}")
        _kwargs = deepcopy(self.connector_contract.kwargs)
        rtn_dict = None
        if connector_type.lower() in ['pickle']:
            return pickle.loads(s3_object['Body'].read())
        raise LookupError(f'The connector type {connector_type} is not currently supported')

    def persist_canonical(self, canonical: dict) -> bool:
        """ persists either the canonical dataset or the YAML contract dictionary"""
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The S3 Source Connector Contract has not been set correctly")
        resource = self.connector_contract.resource
        connector_type = self.connector_contract.connector_type
        bucket = self.connector_contract.location
        _kwargs = self.connector_contract.kwargs
        if connector_type.lower() not in self.supported_types():
            raise ValueError(f"The source type '{connector_type}' is not supported. see supported_types()")
        s3_resource = boto3.resource('s3')
        # create the bucket with the region from the Session
        _kwargs = self.connector_contract.kwargs
        upload_args = {'ACL': 'public-read'}
        if _kwargs.get('ServerSideEncryption') is not None:
            upload_args['ServerSideEncryption'] = _kwargs.get('ServerSideEncryption')
        if _kwargs.get('StorageClass') is not None:
            upload_args['StorageClass'] = _kwargs.get('StorageClass')
        # pickle
        if connector_type.lower() == 'pickle':
            if isinstance(_kwargs, dict) and _kwargs.get('protocol') is not None:
                protocol = _kwargs.get('protocol')
            else:
                protocol = pickle.HIGHEST_PROTOCOL
            pickle_byte_obj = pickle.dumps(canonical, protocol=protocol)
            s3_resource.Object(bucket, resource).put(Body=pickle_byte_obj)
            return True
        # not found
        raise ValueError(f"PandasPersistHandler only supports 'pickle' source type,"
                         f" '{connector_type}' found in Source Contract source-type")

    def remove_canonical(self) -> bool:
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The S3 Source Connector Contract has not been set correctly")
        resource = self.connector_contract.resource
        bucket = self.connector_contract.location
        s3_client = boto3.client('s3')
        response = s3_client.response = s3_client.delete_object(Bucket=bucket, Key=resource)
        if response.get('RequestCharged') is None:
            return False
        return True

    def backup_canonical(self, max_backups=None):
        """ creates a backup of the current source contract resource"""
        if not isinstance(self.connector_contract, ConnectorContract):
            return
        max_backups = max_backups if isinstance(max_backups, int) else 10
        resource = self.connector_contract.resource
        location = self.connector_contract.location
        # _filepath = os.path.join(location, resource)
        # # Check existence of previous versions
        # name, _, ext = _filepath.rpartition('.')
        # for index in range(max_backups):
        #     backup = '%s_%2.2d.%s' % (name, index, ext)
        #     if index > 0:
        #         # No need to backup if file and last version
        #         # are identical
        #         old_backup = '%s_%2.2d.%s' % (name, index - 1, ext)
        #         if not os.path.exists(old_backup):
        #             break
        #         abspath = os.path.abspath(old_backup)
        #
        #         try:
        #             if os.path.isfile(abspath) and filecmp.cmp(abspath, _filepath, shallow=False):
        #                 continue
        #         except OSError:
        #             pass
        #     try:
        #         if not os.path.exists(backup):
        #             shutil.copy(_filepath, backup)
        #     except (OSError, IOError):
        #         pass
        # return



import psycopg2
from ds_foundation.handlers.abstract_handlers import AbstractSourceHandler, ConnectorContract

__author__ = 'Johan Gielstra'


class PostgresSourceHandler(AbstractSourceHandler):
    """ A mongoDB source handler"""

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Hander passing the source_contract dictionary """
        super().__init__(connector_contract)
        self._modified = 0

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['postgresql', 'postgres']

    def load_canonical(self) -> dict:
        """ returns the canonical dataset based on the source contract
            The canonical in this instance is a dictionary that has the headers as the key and then
            the ordered list of values for that header
        """
        conn = None
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Connector Contract is not valid")
        database = self.connector_contract.resource
        host = self.connector_contract.location
        connector_type = self.connector_contract.connector_type
        # jdbc url ?? nicer than individual parameters
        query = self.connector_contract.kwargs.get('query')
        user = self.connector_contract.kwargs.get('user', 'postgres')
        password = self.connector_contract.kwargs.get('password', 'postgres')
        port = self.connector_contract.kwargs.get('port', 5432)
        if connector_type.lower() not in self.supported_types():
            raise ValueError("The source type '{}' is not supported. see supported_types()".format(connector_type))
        try:
            conn = psycopg2.connect(database=database, host=host, port=port, user=user, password=password)
            cur = conn.cursor()
            cur.execute(query)
            colnames = [desc[0] for desc in cur.description]
            rtn_dict = {}
            row = cur.fetchone()  # look into fetchMany versus 1-1
            """
            {
                "col1" = [1,2,3],
                "col2" = ["a","b","c"]
            }
            """
            while row is not None:
                for idx, col in enumerate(colnames):
                    if col not in rtn_dict:
                        rtn_dict[col] = []
                    rtn_dict.get(col).append(row[idx])
                row = cur.fetchone()
            cur.close()
            return rtn_dict
        except (Exception, psycopg2.DatabaseError) as error:
            print(error)
        finally:
            if conn is not None:
                conn.close()
                print('Database connection closed.')

    def get_modified(self) -> [int, float, str]:
        return self._modified



import pandas as pd
from pyhive import hive
from ds_foundation.handlers.abstract_handlers import AbstractSourceHandler, ConnectorContract

__author__ = 'Darryl Oatridge, Neil Pasricha'


class HiveSourceHandler(AbstractSourceHandler):
    """ A Hive source handler"""

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Handler passing the source_contract dictionary """
        super().__init__(connector_contract)
        self._modified = 0

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['hive:python', 'hive:pandas']

    def load_canonical(self) -> [dict, pd.DataFrame]:
        """ returns the canonical dataset based on the source contract
            The canonical in this instance is a dictionary that has the headers as the key and then
            the ordered list of values for that header
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Connector Contract is not valid")
        database = self.connector_contract.resource
        connector_type = self.connector_contract.connector_type
        host = self.connector_contract.location
        user = self.connector_contract.kwargs.get('user')
        password = self.connector_contract.kwargs.get('password')
        auth = self.connector_contract.kwargs.get('auth')
        configuration = self.connector_contract.kwargs.get('configuration')
        kerberos_service_name = self.connector_contract.kwargs.get('kerberos_service_name')
        thrift_transport = self.connector_contract.kwargs.get('thrift_transport')
        query = self.connector_contract.kwargs.get('query')
        host_name, port = host.rsplit(sep=':')
        if connector_type.lower() not in self.supported_types():
            raise ValueError("The source type '{}' is not supported. see supported_types()".format(connector_type))

        conn = hive.Connection(host=host_name, port=port, username=user, password=password, database=database,
                               configuration=configuration, auth=auth, kerberos_service_name=kerberos_service_name,
                               thrift_transport=thrift_transport)
        # return a pandas DataFrame
        if (connector_type.lower().endswith('pandas')):
            return pd.read_sql(query, conn)
        # default return a dictionary
        cursor = conn.cursor()
        cursor.execute(query)

        columns = [i[0] for i in cursor.description]
        rows = cursor.fetchall()
        rtn_dict = {}
        for row in rows:
            for index in range(len(row)):
                if columns[index] not in rtn_dict:
                    rtn_dict[columns[index]] = []
                rtn_dict.get(columns[index]).append(row[index])
        cursor.close()
        return rtn_dict


    def get_modified(self) -> [int, float, str]:
        return self._modified



from pymongo import MongoClient
from ds_foundation.handlers.abstract_handlers import AbstractSourceHandler, ConnectorContract, AbstractPersistHandler

__author__ = 'Darryl Oatridge, Neil Pasricha'


class MongoSourceHandler(AbstractSourceHandler):
    """ A mongoDB source handler"""

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Hander passing the source_contract dictionary """
        super().__init__(connector_contract)
        self._modified = 0

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['mongo']

    def load_canonical(self) -> dict:
        """ returns the canonical dataset based on the source contract
            The canonical in this instance is a dictionary that has the headers as the key and then
            the ordered list of values for that header
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Connector Contract is not valid")
        resource = self.connector_contract.resource
        connector_type = self.connector_contract.connector_type
        location = self.connector_contract.location
        collection = self.connector_contract.kwargs.get('collection')
        filters = self.connector_contract.kwargs.get('filters')
        if connector_type.lower() not in self.supported_types():
            raise ValueError("The source type '{}' is not supported. see supported_types()".format(connector_type))
        client = MongoClient(location)
        db = client[resource]
        stream = db[collection]
        cursor = stream.find(filters)
        rtn_dict = {}
        for line in list(cursor):
            for k, v in line.items():
                if k not in rtn_dict:
                    rtn_dict[k] = []
                rtn_dict.get(k).append(v)
        return rtn_dict

    def get_modified(self) -> [int, float, str]:
        return self._modified


class MongoPersistHandler(AbstractPersistHandler):
    # a mongoDB persist handler

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Hander passing the connector_contract dictionary """
        super().__init__(connector_contract)
        self._modified = 0

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['mongo']

    def get_modified(self) -> [int, float, str]:
        """ returns if the file has been modified"""
        # TODO:
        raise NotImplemented("This is not yet implemented")

    def exists(self) -> bool:
        """ Returns True is the file exists """
        # TODO:
        raise NotImplemented("This is not yet implemented")

    def load_canonical(self) -> dict:
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Connector Contract is not valid")
        resource = self.connector_contract.resource
        connector_type = self.connector_contract.connector_type
        location = self.connector_contract.location
        collection = self.connector_contract.kwargs.get('collection')
        filters = self.connector_contract.kwargs.get('filters')
        if connector_type.lower() not in self.supported_types():
            raise ValueError("The source type '{}' is not supported. see supported_types()".format(connector_type))
        client = MongoClient(location)
        db = client[resource]
        stream = db[collection]
        cursor = stream.find(filters)
        rtn_dict = {}
        for line in list(cursor):
            for k, v in line.items():
                if k not in rtn_dict:
                    rtn_dict[k] = []
                rtn_dict.get(k).append(v)
        return rtn_dict

    def persist_canonical(self, canonical: dict) -> bool:
        """ persists either the canonical dataset"""
        resource = self.connector_contract.resource
        connector_type = self.connector_contract.connector_type
        location = self.connector_contract.location
        collection = self.connector_contract.kwargs.get('collection')
        if connector_type.lower() not in self.supported_types():
            raise ValueError("The source type '{}' is not supported. see supported_types()".format(connector_type))
        client = MongoClient(location)
        db = client[resource]
        stream = db[collection]
        # TODO:
        raise NotImplemented("This is not yet implemented")

    def remove_canonical(self) -> bool:
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Connector Contract has not been set correctly")
        resource = self.connector_contract.resource
        connector_type = self.connector_contract.connector_type
        location = self.connector_contract.location
        collection = self.connector_contract.kwargs.get('collection')
        # TODO:
        raise NotImplemented("This is not yet implemented")

    def backup_canonical(self, max_backups=None):
        """ creates a backup of the current source contract resource"""
        if not isinstance(self.connector_contract, ConnectorContract):
            return
        max_backups = max_backups if isinstance(max_backups, int) else 10
        resource = self.connector_contract.resource
        location = self.connector_contract.location
        # TODO:
        raise NotImplemented("This is not yet implemented")



"""

A setuptools based setup module
with single-source versioning

See:
https://packaging.python.org/en/latest/distributing.html
https://packaging.python.org/guides/single-sourcing-package-version/

"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
import re

here = path.abspath(path.dirname(__file__))


def read(*parts):
    filename = path.join(here, *parts)
    with open(filename, encoding='utf-8') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='discovery-connectors',
    version=find_version('ds_connectors', '__init__.py'),
    description='an concrete implementation of the Discovery Foundation Project to hold a libary of connectors',
    long_description=read('README.rst'),
    url='http://github.com/gigas64/discovery-connectors',
    author='Gigas64',
    author_email='gigas64@opengrass.net',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='Discovery connetors',
    packages=find_packages(exclude=['tests']),
    license='BSD',
    include_package_data=True,
    package_data={},
    python_requires='>=3.6',
    install_requires=[
        'discovery-foundation',
        'pandas>0.25',
        'pymongo',
        'boto3',
        'pyyaml',
        'pyhive',
        'psycopg2'
    ],
    test_suite='tests',
)
